# Reseda version 1.0.0
# ResedaModelParser.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Models
# 4) Nodes and Edges
# 5) Assertion

# 1) Imported modules
# -------------------

from S2MLCore import *

from ResedaModels import *
from ResedaModelLexer import *

# 2) Creator
# ----------

class ResedaModelParser(S2MLModelParser):
  def __init__(self, engine):
    S2MLModelParser.__init__(self, engine)

# 3) Models
# ---------

  def _ParseBlockElement(self, container):
    element = None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token==None:
      return element
    elif token.GetType()==ResedaToken.KEYWORD_NODE:
      element = self._ParseNodeDeclaration(container)
    elif token.GetType()==ResedaToken.KEYWORD_EDGE:
      element = self._ParseEdgeDeclaration(container)
    elif token.GetType()==ResedaToken.KEYWORD_ASSERTION:
      element = self._ParseAssertion(container)
    elif token.GetType()==ResedaToken.KEYWORD_PARAMETER:
      element = self._ParseParameter(container)
    elif token.GetType()==ResedaToken.KEYWORD_BLOCK:
      element = self._ParseBlock(container)
    elif token.GetType()==ResedaToken.KEYWORD_EXTENDS:
      element = self._ParseExtendsDirective(container)
    elif token.GetType()==ResedaToken.KEYWORD_EMBEDS:
      element = self._ParseEmbedsDirective(container)
    elif token.GetType()==ResedaToken.KEYWORD_ATTRIBUTE:
      element = self._ParseAttributeDirective(container)
    elif token.GetType()==ResedaToken.KEYWORD_CLONES:
      element = self._ParseClonesDirective(container)
    else:
      instantiatedElementIdentifier = self._ParseIdentifier()
      if instantiatedElementIdentifier==None:
        return None
      instantiatedElement = self._model.LookForIdentifiedElement(instantiatedElementIdentifier)
      if instantiatedElement==None:
        self._RaiseParseError(token, "Undeclared domain or class %s" % instantiatedElementIdentifier)
        return None
      self._currentTokenIndex -= 1
      element = self._ParseClassInstance(container)
    return element


# 4) Nodes and Edges
# ------------------

  def _ParseNodeDeclaration(self, container):
    token = self._ParseGivenToken(ResedaToken.KEYWORD_NODE)
    if token==None:
      return None
    nodes = []
    while True:
      path = self._ParsePath()
      if path==None:
        return None
      node = self._model.NewNode(path, container)
      nodes.append(node)
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==ResedaToken.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==ResedaToken.TOKEN_LPAR:
      nodes = self._ParseAttributes(nodes)
      if nodes==None:
        return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    return nodes

  def _ParseEdgeDeclaration(self, container):
    token = self._ParseGivenToken(ResedaToken.KEYWORD_EDGE)
    if token==None:
      return None
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token==None or (token.GetType()!=ResedaToken.TOKEN_IDENTIFIER and token.GetType()!=ResedaToken.TOKEN_PATH):
        break
      edge = self._ParseEdge(container)
      if edge==None:
        # self._RecoverError(ResedaToken.KEYWORD_END)
        return None
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==ResedaToken.TOKEN_COMMA:
        self._currentTokenIndex += 1
      else:
        break
    token = self._ParseGivenToken(ResedaToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    return container

  def _ParseEdge(self, container):
    token = self._lexer.GetToken(self._currentTokenIndex)
    path = self._ParsePath()
    if path==None:
      return None
    attributes = self._ParseAttributeList()
    if attributes==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_COLON)
    if token==None:
      return None
    sourcePortPath = self._ParsePath()
    if sourcePortPath==None:
      return None
    tokenArrow = self._lexer.GetToken(self._currentTokenIndex)
    if tokenArrow.GetType()==ResedaToken.TOKEN_RIGHT_ARROW:
      self._currentTokenIndex += 1
    elif tokenArrow.GetType()==ResedaToken.TOKEN_LEFT_RIGHT_ARROW:
      self._currentTokenIndex += 1
    else:
      self._RaiseParseError(tokenArrow, "An arrow (-> or <->) was expected")
      return None
    targetPortPath = self._ParsePath()
    if targetPortPath==None:
      return None
    if tokenArrow.GetType()==ResedaToken.TOKEN_RIGHT_ARROW:
      edge = self._model.NewEdge(path, sourcePortPath, targetPortPath, container)
    else:
      edge = self._model.NewArc(path, sourcePortPath, targetPortPath, container)
    for (name, expression) in attributes:
      edge.SetAttribute(name, expression)
    return edge

# 5) Assertion
# ------------

  def _ParseAssertion(self, container):
    token = self._ParseGivenToken(ResedaToken.KEYWORD_ASSERTION)
    if token==None:
      return None
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=ResedaToken.KEYWORD_SET:
        break
      variable = self._ParseVariableDeclaration(container)
      if variable==None:
        return None
    return container

  def _ParseVariableDeclaration(self, container):
    token = self._ParseGivenToken(ResedaToken.KEYWORD_SET)
    if token==None:
      return None
    path = self._ParsePath()
    if path==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_EQUAL)
    if token==None:
      return None
    definition = self._ParseSetExpression()
    if definition==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_SEMICOLON)
    if token==None:
      return None
    variable = self._model.NewVariable(path, container)
    variable.SetDefinition(definition)
    return variable

  def _ParseSetExpression(self):
    expression = self._ParseSetExpression3()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=ResedaToken.TOKEN_VEE:
      return expression
    children = [expression]
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=ResedaToken.TOKEN_VEE:
        break
      self._currentTokenIndex += 1
      expression = self._ParseSetExpression3()
      if expression==None:
        return None
      children.append(expression)
    expression = ResedaSetExpression.NewOperation(ResedaSetExpression.UNION, children)
    return expression

  def _ParseSetExpression3(self):
    expression = self._ParseSetExpression2()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=ResedaToken.TOKEN_WEDGE:
      return expression
    children = [expression]
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=ResedaToken.TOKEN_WEDGE:
        break
      self._currentTokenIndex += 1
      expression = self._ParseSetExpression2()
      if expression==None:
        return None
      children.append(expression)
    expression = ResedaSetExpression.NewOperation(ResedaSetExpression.INTERSECTION, children)
    return expression

  def _ParseSetExpression2(self):
    expression = self._ParseSetExpression1()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=ResedaToken.TOKEN_SLASH:
      return expression
    children = [expression]
    while True:
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()!=ResedaToken.TOKEN_SLASH:
        break
      self._currentTokenIndex += 1
      expression = self._ParseSetExpression1()
      if expression==None:
        return None
      children.append(expression)
    expression = ResedaSetExpression.NewOperation(ResedaSetExpression.SET_DIFFERENCE, children)
    return expression

  def _ParseSetExpression1(self):
    expression = self._ParseSetExpression0()
    if expression==None:
      return None
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()!=ResedaToken.TOKEN_COLON:
      return expression
    self._currentTokenIndex += 1
    testOnAttributes = self._ParseStochasticExpression()
    if testOnAttributes==None:
      return None
    expression = ResedaSetExpression.NewSelection(expression, testOnAttributes)
    return expression

  def _ParseSetExpression0(self):
    token = self._lexer.GetToken(self._currentTokenIndex)
    if token.GetType()==ResedaToken.TOKEN_IDENTIFIER or token.GetType()==ResedaToken.TOKEN_PATH:
      path = self._ParsePath()
      return ResedaSetExpression.NewElementReference(path)
    if token.GetType()==ResedaToken.TOKEN_LBRACE:
      token = self._ParseGivenToken(ResedaToken.TOKEN_LBRACE)
      expression = ResedaSetExpression.NewExtensionalSet()
      token = self._lexer.GetToken(self._currentTokenIndex)
      if token.GetType()==ResedaToken.TOKEN_RBRACE:
        token = self._ParseGivenToken(ResedaToken.TOKEN_RBRACE)
        return expression
      while True:
        child = self._ParseSetExpression()
        if child==None:
          return None
        expression.AddChild(child)
        token = self._lexer.GetToken(self._currentTokenIndex)
        if token.GetType()!=ResedaToken.TOKEN_COMMA:
          break
        token = self._ParseGivenToken(ResedaToken.TOKEN_COMMA)
      token = self._ParseGivenToken(ResedaToken.TOKEN_RBRACE)
      if token==None:
        return None
      return expression
    if token.GetType()==ResedaToken.TOKEN_LPAR:
      token = self._ParseGivenToken(ResedaToken.TOKEN_LPAR)
      expression = self._ParseSetExpression()
      if expression==None:
        return None
      token = self._ParseGivenToken(ResedaToken.TOKEN_RPAR)
      if token==None:
        return None
      return expression
    if token.GetType()==ResedaToken.KEYWORD_PARENT_BLOCKS:
      return self.ParseUnaryOperation(ResedaSetExpression.PARENT_BLOCKS)
    if token.GetType()==ResedaToken.KEYWORD_ANCESTOR_BLOCKS:
      return self.ParseUnaryOperation(ResedaSetExpression.ANCESTOR_BLOCKS)
    if token.GetType()==ResedaToken.KEYWORD_CHILD_BLOCKS:
      return self.ParseUnaryOperation(ResedaSetExpression.CHILD_BLOCKS)
    if token.GetType()==ResedaToken.KEYWORD_DESCENDANT_BLOCKS:
      return self.ParseUnaryOperation(ResedaSetExpression.DESCENDANT_BLOCKS)
    if token.GetType()==ResedaToken.KEYWORD_CHILD_NODES:
      return self.ParseUnaryOperation(ResedaSetExpression.CHILD_NODES)
    if token.GetType()==ResedaToken.KEYWORD_DESCENDANT_NODES:
      return self.ParseUnaryOperation(ResedaSetExpression.DESCENDANT_NODES)
    if token.GetType()==ResedaToken.KEYWORD_CHILD_EDGES:
      return self.ParseUnaryOperation(ResedaSetExpression.CHILD_EDGES)
    if token.GetType()==ResedaToken.KEYWORD_DESCENDANT_EDGES:
      return self.ParseUnaryOperation(ResedaSetExpression.DESCENDANT_EDGES)
    if token.GetType()==ResedaToken.KEYWORD_SOURCE_NODES:
      return self.ParseUnaryOperation(ResedaSetExpression.SOURCE_NODES)
    if token.GetType()==ResedaToken.KEYWORD_TARGET_NODES:
      return self.ParseUnaryOperation(ResedaSetExpression.TARGET_NODES)
    if token.GetType()==ResedaToken.KEYWORD_IN_EDGES:
      return self.ParseUnaryOperation(ResedaSetExpression.IN_EDGES)
    if token.GetType()==ResedaToken.KEYWORD_OUT_EDGES:
      return self.ParseUnaryOperation(ResedaSetExpression.OUT_EDGES)
    if token.GetType()==ResedaToken.KEYWORD_REACHABLE_NODES:
      return self.ParseReachableNodesOperation(ResedaSetExpression.REACHABLE_NODES)
    if token.GetType()==ResedaToken.KEYWORD_COREACHABLE_NODES:
      return self.ParseReachableNodesOperation(ResedaSetExpression.COREACHABLE_NODES)
    self._RaiseParseError(token, "Invalid set expression")
    return None

  def ParseUnaryOperation(self, operator):
    self._currentTokenIndex += 1
    token = self._ParseGivenToken(ResedaToken.TOKEN_LPAR)
    if token==None:
      return None
    child = self._ParseSetExpression()
    if child==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_RPAR)
    if token==None:
      return None
    return ResedaSetExpression.NewOperation(operator, [child])

  def ParseReachableNodesOperation(self, operator):
    self._currentTokenIndex += 1
    token = self._ParseGivenToken(ResedaToken.TOKEN_LPAR)
    if token==None:
      return None
    child = self._ParseSetExpression()
    if child==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_COMMA)
    if token==None:
      return None
    testOnAttributesOfNodes = self._ParseStochasticExpression()
    if testOnAttributesOfNodes==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_COMMA)
    if token==None:
      return None
    testOnAttributesOfEdges = self._ParseStochasticExpression()
    if testOnAttributesOfEdges==None:
      return None
    token = self._ParseGivenToken(ResedaToken.TOKEN_RPAR)
    if token==None:
      return None
    return ResedaSetExpression.NewReachableNodesOperation(operator, child, testOnAttributesOfNodes, testOnAttributesOfEdges)


