# Reseda version 1.0.0
# ResedaModelPrinter.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda, a pedagogical stochastic simulator.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) S2ML
# 4) Nodes and edges
# 5) Sets
# 6) Values of parameters and attributes

# 1) Imported modules
# -------------------

from S2MLCore import *
from ResedaModels import *
from ResedaToken import *
from MusetCore import *

# 2) Creator
# ----------

class ResedaModelPrinter(S2MLModelPrinter):
  S2ML_SECTION = 0
  ASSERTION_SECTION = 1

  def __init__(self, engine):
    S2MLModelPrinter.__init__(self, engine)
    self._section = ResedaModelPrinter.S2ML_SECTION

# 3) Models
# ---------

  def PrintElement(self, element, output):
    if element.GetType()==ResedaType.NODE:
      self._UpdateSection(ResedaModelPrinter.S2ML_SECTION, output)
      self.PrintNode(element, output)
    elif element.GetType()==ResedaType.EDGE:
      self._UpdateSection(ResedaModelPrinter.S2ML_SECTION, output)
      self.PrintEdge(element, output)
    elif element.GetType()==ResedaType.ARC:
      self._UpdateSection(ResedaModelPrinter.S2ML_SECTION, output)
      self.PrintEdge(element, output)
    elif element.GetType()==ResedaType.VARIABLE:
      self._UpdateSection(ResedaModelPrinter.ASSERTION_SECTION, output)
      self.PrintVariable(element, output)
    else:
      self._UpdateSection(ResedaModelPrinter.S2ML_SECTION, output)
      S2MLModelPrinter.PrintElement(self, element, output)

  def _UpdateSection(self, section, output):
    if self._section==ResedaModelPrinter.ASSERTION_SECTION and section==ResedaModelPrinter.S2ML_SECTION:
      self._section = ResedaModelPrinter.S2ML_SECTION
      self._depth -= 1
    elif self._section==ResedaModelPrinter.S2ML_SECTION and section==ResedaModelPrinter.ASSERTION_SECTION:
      self._section = ResedaModelPrinter.ASSERTION_SECTION
      self.PrintIndentation(output)
      self.PrintKeyword(ResedaToken.KEYWORD_ASSERTION, output)
      output.write("\n")
      self._depth += 1

  def _PrintBlockBody(self, block, output):
    self._section = ResedaModelPrinter.S2ML_SECTION
    self._depth += 1
    for element in block.GetElements():
      self.PrintElement(element, output)
    self._UpdateSection(ResedaModelPrinter.S2ML_SECTION, output)
    self._depth -= 1

# 4) Nodes and edges
# ------------------

  def PrintNode(self, node, output):
    self.PrintIndentation(output)
    self.PrintKeyword(ResedaToken.KEYWORD_NODE, output)
    output.write(" ")
    self.PrintPath(node.GetPath(), output)
    self.PrintAttributes(node, output)
    output.write(";\n")

  def PrintEdge(self, edge, output):
    self.PrintIndentation(output)
    self.PrintKeyword(ResedaToken.KEYWORD_EDGE, output)
    output.write(" ")
    self.PrintPath(edge.GetPath(), output)
    self.PrintAttributes(edge, output)
    output.write(": ")
    nodePath = edge.GetSourceNodePath()
    self.PrintPath(nodePath, output)
    if edge.GetType()==ResedaType.EDGE:
      output.write(" -> ")
    else:
      output.write(" <-> ")
    nodePath = edge.GetTargetNodePath()
    self.PrintPath(nodePath, output)
    output.write(";\n")

# 5) Sets
# -------

  def PrintVariable(self, variable, output):
    self.PrintIndentation(output)
    self.PrintKeyword(ResedaToken.KEYWORD_SET, output)
    output.write(" ")
    self.PrintPath(variable.GetPath(), output)
    output.write(" = ")
    self.PrintSetExpression(variable.GetDefinition(), output)
    output.write(";\n")

  def PrintSetExpression(self, expression, output):
    if expression.GetType()==ResedaSetExpression.ELEMENT_REFERENCE:
      path = expression.GetReferencePath()
      self.PrintPath(path, output)
    elif expression.GetType()==ResedaSetExpression.EXTENSIONAL_SET:
      output.write("{")
      self.PrintChildrenOfSetExpression(expression, ", ", output)
      output.write("}")
    elif expression.GetType()==ResedaSetExpression.UNION:
      self.PrintChildrenOfSetExpression(expression, " \\/ ", output)
    elif expression.GetType()==ResedaSetExpression.INTERSECTION:
      self.PrintChildrenOfSetExpression(expression, " /\\ ", output)
    elif expression.GetType()==ResedaSetExpression.SET_DIFFERENCE:
      self.PrintChildrenOfSetExpression(expression, " / ", output)
    elif expression.GetType()==ResedaSetExpression.SELECTION:
      child = expression.GetChild(0)
      self.PrintSetExpression(child, output)
      output.write(": ")
      testOnAttributes = expression.GetTestOnAttributes(0)
      self.PrintStochasticExpression(testOnAttributes, output)
    elif expression.GetType()==ResedaSetExpression.PARENT_BLOCKS:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_PARENT_BLOCKS, output)
    elif expression.GetType()==ResedaSetExpression.ANCESTOR_BLOCKS:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_ANCESTOR_BLOCKS, output)
    elif expression.GetType()==ResedaSetExpression.CHILD_BLOCKS:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_CHILD_BLOCKS, output)
    elif expression.GetType()==ResedaSetExpression.DESCENDANT_BLOCKS:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_DESCENDANT_BLOCKS, output)
    elif expression.GetType()==ResedaSetExpression.CHILD_NODES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_CHILD_NODES, output)
    elif expression.GetType()==ResedaSetExpression.DESCENDANT_NODES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_DESCENDANT_NODES, output)
    elif expression.GetType()==ResedaSetExpression.CHILD_EDGES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_CHILD_EDGES, output)
    elif expression.GetType()==ResedaSetExpression.DESCENDANT_EDGES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_DESCENDANT_EDGES, output)
    elif expression.GetType()==ResedaSetExpression.SOURCE_NODES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_SOURCE_NODES, output)
    elif expression.GetType()==ResedaSetExpression.TARGET_NODES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_TARGET_NODES, output)
    elif expression.GetType()==ResedaSetExpression.IN_EDGES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_IN_EDGES, output)
    elif expression.GetType()==ResedaSetExpression.OUT_EDGES:
      self.PrintSetOperator(expression, ResedaToken.KEYWORD_OUT_EDGES, output)
    elif expression.GetType()==ResedaSetExpression.REACHABLE_NODES:
      self.PrintReachableNodesOperator(expression, ResedaToken.KEYWORD_REACHABLE_NODES, output)
    elif expression.GetType()==ResedaSetExpression.COREACHABLE_NODES:
      self.PrintReachableNodesOperator(expression, ResedaToken.KEYWORD_COREACHABLE_NODES, output)

  def PrintChildrenOfSetExpression(self, expression, separator, output):
    first = True
    for child in expression.GetChildren():
      if first:
        first = False
      else:
        output.write(separator)
      if child.GetPriority()>=expression.GetPriority() and child.GetPriority()!=0:
        output.write("(")
      self.PrintSetExpression(child, output)
      if child.GetPriority()>=expression.GetPriority() and child.GetPriority()!=0:
        output.write(")")

  def PrintSetOperator(self, expression, functor, output):
    self.PrintKeyword(functor, output)
    output.write("(")
    child = expression.GetChild(0)
    self.PrintSetExpression(child, output)
    output.write(")")

  def PrintReachableNodesOperator(self, expression, functor, output):
    self.PrintKeyword(functor, output)
    output.write("(")
    child = expression.GetChild(0)
    self.PrintSetExpression(child, output)
    output.write(", ")
    testOnAttributes = expression.GetTestOnAttributes(0)
    self.PrintStochasticExpression(testOnAttributes, output)
    output.write(", ")
    testOnAttributes = expression.GetTestOnAttributes(1)
    self.PrintStochasticExpression(testOnAttributes, output)
    output.write(")")


# 6) Values of parameters and attributes
# --------------------------------------

  def PrintValuesOfElements(self, block, printParameters, printAttributes, printProperties, output):
    for element in block.GetElements():
      if element.GetType()==ResedaType.BLOCK:
        self.PrintValuesOfElements(element, printParameters, printAttributes, printProperties, output)
      elif element.GetType()==ResedaType.PARAMETER:
        if printParameters:
          self.PrintValueOfParameter(element, output)
      elif element.GetType()==ResedaType.NODE:
        if printAttributes:
          self.PrintValuesOfAttributesOfElement(element, output)
      elif element.GetType()==ResedaType.EDGE:
        if printAttributes:
          self.PrintValuesOfAttributesOfElement(element, output)
      elif element.GetType()==ResedaType.ARC:
        if printAttributes:
          self.PrintValuesOfAttributesOfElement(element, output)
      elif element.GetType()==ResedaType.VARIABLE:
        if printProperties:
          self.PrintValueOfVariable(element, output)

  def PrintValueOfParameter(self, parameter, output):
    path = parameter.GetContextualizedPath()
    definition = parameter.GetDefinition()
    value = definition.GetValue()
    self.PrintPath(path, output)
    output.write(" = ")
    self.PrintConstant(value, output)
    output.write("\n")

  def PrintValuesOfAttributesOfElement(self, element, output):
    path = element.GetContextualizedPath()
    for name in element.GetAttributeNames():
      expression = element.GetAttribute(name)
      value = expression.GetValue()
      self.PrintPath(path, output)
      output.write(".{0:s}".format(name))
      output.write(" = ")
      self.PrintConstant(value, output)
      output.write("\n")

  def PrintValueOfVariable(self, variable, output):
    path = variable.GetContextualizedPath()
    self.PrintPath(path, output)
    output.write(" = ")
    value = variable.GetValue()
    self.PrintMusetElement(value, output)
    output.write(";\n")

  def PrintMusetElement(self, musetElement, output):
    if musetElement.GetType()==MusetElement.NONE:
      output.write("none")
    elif musetElement.GetType()==MusetElement.CONSTANT:
      constant = musetElement.GetConstant()
      self.PrintConstant(constant, output)
    elif musetElement.GetType()==MusetElement.BLOCK:
      block = musetElement.GetBlock()
      path = block.GetContextualizedPath()
      if path.IsEmpty():
        self.PrintKeyword(ResedaToken.KEYWORD_MAIN, output)
      else:
        self.PrintPath(path, output)
    elif musetElement.GetType()==MusetElement.NODE:
      node = musetElement.GetNode()
      path = node.GetContextualizedPath()
      self.PrintPath(path, output)
    elif musetElement.GetType()==MusetElement.EDGE:
      edge = musetElement.GetEdge()
      path = edge.GetContextualizedPath()
      self.PrintPath(path, output)
    elif musetElement.GetType()==MusetElement.SET:
      self.PrintMusetSet(musetElement, output)

  def PrintMusetSet(self, musetSet, output):
    output.write("{")
    first = True
    for musetElement in musetSet.GetElements():
      if first:
        first = False
      else:
        output.write(", ")
      self.PrintMusetElement(musetElement, output)
    output.write("}")
