# Reseda version 1.0.0
# ResedaSetExpression.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creators
# 3) Accessors
# 4) Children
# 5) Tests on attributes

# 1) Imported modules
# -------------------

from S2MLCore import *

# 2) Creator
# ----------

class ResedaSetExpression:
  ELEMENT_REFERENCE = 1
  EXTENSIONAL_SET = 2
  UNION = 12
  INTERSECTION = 13
  SET_DIFFERENCE = 14
  SELECTION = 15
  PARENT_BLOCKS = 21
  ANCESTOR_BLOCKS = 22
  CHILD_BLOCKS = 23
  DESCENDANT_BLOCKS = 24
  CHILD_NODES = 25
  DESCENDANT_NODES = 26
  CHILD_EDGES = 27
  DESCENDANT_EDGES = 28
  SOURCE_NODES = 31
  TARGET_NODES = 32
  IN_EDGES = 33
  OUT_EDGES = 34
  REACHABLE_NODES = 5
  COREACHABLE_NODES = 36

  def __init__(self, type):
    self._type = type
    self._referencePath = None
    self._children = []
    self._testsOnAttributes = []

  @staticmethod
  def NewElementReference(path):
    self = ResedaSetExpression(ResedaSetExpression.ELEMENT_REFERENCE)
    self._referencePath = path
    return self

  @staticmethod
  def NewExtensionalSet():
    self = ResedaSetExpression(ResedaSetExpression.EXTENSIONAL_SET)
    return self

  @staticmethod
  def NewSelection(child, testOnAttributes):
    self = ResedaSetExpression(ResedaSetExpression.SELECTION)
    self.AddChild(child)
    self.AddTestOnAttributes(testOnAttributes)
    return self

  @staticmethod
  def NewOperation(type, children):
    self = ResedaSetExpression(type)
    self.AddChildren(children)
    return self

  @staticmethod
  def NewReachableNodesOperation(type, child, testOnAttributesOfNodes, testOnAttributesOfEdges):
    self = ResedaSetExpression(type)
    self.AddChild(child)
    self.AddTestOnAttributes(testOnAttributesOfNodes)
    self.AddTestOnAttributes(testOnAttributesOfEdges)
    return self

# 3) Accessors
# ------------

  def GetType(self):
    return self._type

  def IsElementReference(self):
    return self._type==ResedaSetExpression.ELEMENT_REFERENCE

  def GetReferencePath(self):
    if self._type!=ResedaSetExpression.ELEMENT_REFERENCE:
      return None
    return self._referencePath

  def GetTestOnAttributes(self):
    return self._testOnAttributes

  def GetPriority(self):
    if self._type==ResedaSetExpression.UNION:
      return 4
    if self._type==ResedaSetExpression.INTERSECTION:
      return 3
    if self._type==ResedaSetExpression.SET_DIFFERENCE:
      return 2
    if self._type==ResedaSetExpression.SELECTION:
      return 1
    return 0

# 4) Children
# -----------

  def GetNumberOfChildren(self):
    return len(self._children)

  def GetChildren(self):
    return self._children

  def GetChild(self, index):
    if index<0 or index>=self.GetNumberOfChildren():
      return None
    return self._children[index]

  def AddChild(self, child):
    self._children.append(child)

  def AddChildren(self, children):
    for child in children:
       self._children.append(child)

# 5) Tests on attributes
# ----------------------

  def GetNumberOfTestOnAttributes(self):
    return len(self._testsOnAttributes)

  def GetTestOnAttributes(self):
    return self._testsOnAttributes

  def GetTestOnAttributes(self, index):
    if index<0 or index>=self.GetNumberOfTestOnAttributes():
      return None
    return self._testsOnAttributes[index]

  def AddTestOnAttributes(self, testOnAttributes):
    self._testsOnAttributes.append(testOnAttributes)

  def AddTestsOnAttributes(self, testsOnAttributes):
    for testOnAttributes in testsOnAttributes:
       self._testsOnAttributes.append(testOnAttributes)

