# Reseda version 1.0.0
# ResedaMaximumFlowCalculator.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda, a pedagogical implementation of graphs as a modeling tool.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Ford-Fulkerson algorithm

# 1) Imported modules
# -------------------

import string
import sys
import re
import math

from S2MLCore import *
from ResedaModels import *

# 2) Creator
# ----------

class ResedaMaximumFlowCalculator:
  CAPACITY = "@CAPACITY"
  FLOW = "@FLOW"
  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._expressionManager = self._engine.GetExpressionManager()
    self._targetBlock = None
    self._sourceNode = None
    self._targetNode = None
    self._augmentingPath = []
    self._augmentingCapacity = math.inf
    self._matrix = None

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) Ford-Fulkerson algorithm
# ---------------------------

  def ComputeMaximumFlowWithFordFulkersonAlgorithm(self, block, sourceNodeName, targetNodeName, capacityName):
    self._targetBlock = block
    if self._targetBlock==None:
      self.RaiseError("No target block")
      return None
    sourceNodePath = S2MLPath.FromString(sourceNodeName)
    self._sourceNode = self._targetBlock.AccessNode(sourceNodePath)
    if self._sourceNode==None:
      self.RaiseError("No node %s" % sourceNodeName)
      return None
    targetNodePath = S2MLPath.FromString(targetNodeName)
    self._targetNode = self._targetBlock.AccessNode(targetNodePath)
    if self._targetNode==None:
      self.RaiseError("No node %s" % targetNodeName)
      return None
    error = self.BuildAndInitializeGraphToComputeMaximumFlow(capacityName)
    if error:
      return None
    self.CalculateMaximumFlow()
    self.CalculateFlowInNodes()
    matrix = self.CreateResultMatrixForMaximumFlow()
    self._targetBlock.ResetAttributesOfElements(ResedaModel.IsEdge, ResedaMaximumFlowCalculator.CAPACITY, False)
    self._targetBlock.ResetAttributesOfElements(ResedaModel.IsNode, ResedaMaximumFlowCalculator.FLOW, False)
    self._targetBlock.ResetAttributesOfElements(ResedaModel.IsEdge, ResedaMaximumFlowCalculator.FLOW, False)
    return matrix

  def BuildAndInitializeGraphToComputeMaximumFlow(self, capacityName):
    self._targetBlock.BuildGraph()
    self._expressionManager.ResetValuesOfParameters(self._targetBlock)
    for edge in self._targetBlock.GetEdges():
      expression = edge.GetAttribute(capacityName)
      if expression==None:
        self.RaiseError("Missing attribute %s in edge %s." % (capacityName, edge.GetIdentifier()))
        return 1
      else:
        value = self._expressionManager.CalculateValueOfExpression(expression)
        if not value.IsNumber():
          self.RaiseError("Attribute %s of edge %s is not a number" % (capacityName, edge.GetIdentifier()))
          return 1
        capacity = value.ToFloat()
        if capacity < 0:
          self.RaiseError("Attribute %s of edge %s has a negative value" % (capacityName, edge.GetIdentifier()))
          return 1
        edge.SetAttribute(ResedaMaximumFlowCalculator.CAPACITY, capacity)
        edge.SetAttribute(ResedaMaximumFlowCalculator.FLOW, 0.0)
    return 0

  def CalculateMaximumFlow(self):
    if self._sourceNode==self._targetNode:
      return
    while True:
      self._augmentingPath = []
      self._augmentingCapacity = math.inf
      found = self.SelectAugmentingPath(self._sourceNode)
      if not found:
        break
      self.InsertAugmentingPath()

  def SelectAugmentingPath(self, node):
    if node==self._targetNode:
      return True
    if node.GetFlag()!=0:
      return False
    node.SetFlag(1)
    for edge in node.GetOutEdges():
      if edge.GetSourceNode()==node:
        edgeCapacity = edge.GetAttribute(ResedaMaximumFlowCalculator.CAPACITY)
        edgeFlow = edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW)
        edgeResidualCapacity = edgeCapacity - edgeFlow
        if edgeResidualCapacity>0:
          previousAugmentingCapacity = self._augmentingCapacity
          if edgeResidualCapacity<self._augmentingCapacity:
            self._augmentingCapacity = edgeResidualCapacity
          oppositeNode = edge.GetTargetNode()
          self._augmentingPath.append(edge)
          found = self.SelectAugmentingPath(oppositeNode)
          if found:
            node.SetFlag(0)
            return True
          self._augmentingPath.pop()
          self._augmentingCapacity = previousAugmentingCapacity
    for edge in node.GetInEdges():
      if edge.GetTargetNode()==node:
        edgeCapacity = edge.GetAttribute(ResedaMaximumFlowCalculator.CAPACITY)
        edgeFlow = edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW)
        if edge.IsArc():
          edgeResidualCapacity = edgeFlow - edgeCapacity
        else:
          edgeResidualCapacity = -edgeFlow
        if edgeResidualCapacity<0:
          previousAugmentingCapacity = self._augmentingCapacity
          if -edgeResidualCapacity<self._augmentingCapacity:
            self._augmentingCapacity = -edgeResidualCapacity
          oppositeNode = edge.GetSourceNode()
          self._augmentingPath.append(edge)
          found = self.SelectAugmentingPath(oppositeNode)
          if found:
            node.SetFlag(0)
            return True
          self._augmentingPath.pop()
          self._augmentingCapacity = previousAugmentingCapacity
    node.SetFlag(0)
    return False

  def CalculateFlowInNodes(self):
    for node in self._targetBlock.GetNodes():
      node.SetAttribute(ResedaMaximumFlowCalculator.FLOW, 0.0)
    for edge in self._targetBlock.GetEdges():
      edgeFlow = edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW)
      if edge.IsArc() and edgeFlow<0:
        node = edge.GetSourceNode()
      else:
        node = edge.GetTargetNode()
      node.SetAttribute(ResedaMaximumFlowCalculator.FLOW, node.GetAttribute(ResedaMaximumFlowCalculator.FLOW) + edgeFlow)
    self._sourceNode.SetAttribute(ResedaMaximumFlowCalculator.FLOW, self._targetNode.GetAttribute(ResedaMaximumFlowCalculator.FLOW))

  def InsertAugmentingPath(self):
    node = self._sourceNode
    for edge in self._augmentingPath:
      if node==edge.GetSourceNode():
        node = edge.GetTargetNode()
        edge.SetAttribute(ResedaMaximumFlowCalculator.FLOW, edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW) + self._augmentingCapacity)
      else:
        node = edge.GetSourceNode()
        edge.SetAttribute(ResedaMaximumFlowCalculator.FLOW, edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW) - self._augmentingCapacity)

  def CreateResultMatrixForMaximumFlow(self):
    matrix = S2MLMatrix()
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Model"))
    row.Put(1, S2MLConstant.NewString(self._targetBlock.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("source"))
    row.Put(1, S2MLConstant.NewString(self._sourceNode.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("target"))
    row.Put(1, S2MLConstant.NewString(self._targetNode.GetIdentifier()))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("node"))
    row.Put(1, S2MLConstant.NewString("flow"))
    for node in self._targetBlock.GetNodes():
      row = matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString(node.GetIdentifier()))
      row.Put(1, S2MLConstant.NewReal(node.GetAttribute(ResedaMaximumFlowCalculator.FLOW)))
    row = matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("edge"))
    row.Put(1, S2MLConstant.NewString("source"))
    row.Put(2, S2MLConstant.NewString("target"))
    row.Put(3, S2MLConstant.NewString("capacity"))
    row.Put(4, S2MLConstant.NewString("flow"))
    for edge in self._targetBlock.GetEdges():
      row = matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString(edge.GetIdentifier()))
      row.Put(1, S2MLConstant.NewString(edge.GetSourceNode().GetIdentifier()))
      row.Put(2, S2MLConstant.NewString(edge.GetTargetNode().GetIdentifier()))
      row.Put(3, S2MLConstant.NewReal(edge.GetAttribute(ResedaMaximumFlowCalculator.CAPACITY)))
      row.Put(4, S2MLConstant.NewReal(edge.GetAttribute(ResedaMaximumFlowCalculator.FLOW)))
    return matrix


