# Reseda version 1.0.0
# ResedaModelFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imnodeed modules
# 2) Creator
# 3) Instantiation of elements
# 4) Flattening of elements

# 1) Imnodeed modules
# -------------------

from S2MLCore import *

from ResedaModels import *
from ResedaModelPrinter import *

# 2) Creator
# ----------

class ResedaModelFlattener(S2MLModelFlattener):
  def __init__(self, engine):
    S2MLModelFlattener.__init__(self, engine)

# 3) Instantiation of elements
# ----------------------------

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==ResedaType.NODE:
      status = self._InstantiateNode(action)
    elif sourceElement.GetType()==ResedaType.EDGE:
      status = self._InstantiateEdge(action)
    elif sourceElement.GetType()==ResedaType.ARC:
      status = self._InstantiateEdge(action)
    elif sourceElement.GetType()==ResedaType.VARIABLE:
      status = self._InstantiateVariable(action)
    else:
      status = S2MLModelFlattener.InstantiateElement(self, action)
    return status

  def _InstantiateNode(self, action):
    sourceNode = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceNode.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetNode = self._targetModel.NewNode(targetPath, targetContainer)
      else:
        absoluteSourcePath = sourceNode.GetAbsolutePath()
        self.RaiseError("Node {0:s} declared outside of its parent block\n".format(absoluteSourcePath.ToString()))
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==ResedaType.NODE:
      targetNode = targetElement
    else:
      absoluteSourcePath = sourceNode.GetAbsolutePath()
      self.RaiseError("Type mismatch for node {0:s}\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    actualTargetContainer = targetNode.GetContainer()
    for name in sourceNode.GetAttributeNames():
      sourceExpression = sourceNode.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetNode.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateEdge(self, action):
    sourceEdge = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    # 1. Path
    targetPath = sourceEdge.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if not targetPath.IsIdentifier():
        absoluteSourcePath = sourceEdge.GetAbsolutePath()
        self.RaiseError("Edge {0:s} declared outside of its parent block\n".format(absoluteSourcePath.ToString()))
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()!=ResedaType.EDGE and targetElement.GetType()!=ResedaType.ARC:
      absoluteSourcePath = sourceNode.GetAbsolutePath()
      self.RaiseError("Type mismatch for edge {0:s}\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    # 2. Source node path
    targetSourceNodePath = sourceEdge.GetSourceNodePath()
    targetSourceNodeElement = targetContainer.AccessNamedElement(targetSourceNodePath)
    if targetSourceNodeElement==None:
      absoluteSourcePath = sourceEdge.GetAbsolutePath()
      self.RaiseError("In edge {0:s}, undeclared source node {1:s}\n".format(absoluteSourcePath.ToString(), targetSourceNodePath.ToString()))
      return None
    elif targetSourceNodeElement.GetType()==ResedaType.NODE:
      targetSourceNode = targetSourceNodeElement
    else:
      absoluteSourcePath = sourceEdge.GetAbsolutePath()
      self.RaiseError("In edge {0:s}, type mismatch for source node {1:s}\n".format(absoluteSourcePath.ToString(), targetSourceNodePath.ToString()))
      return None
    # 3. Target node path
    targetTargetNodePath = sourceEdge.GetTargetNodePath()
    targetTargetNodeElement = targetContainer.AccessNamedElement(targetTargetNodePath)
    if targetTargetNodeElement==None:
      absoluteSourcePath = sourceEdge.GetAbsolutePath()
      self.RaiseError("In connection {0:s}, undeclared target node {1:s}\n".format(absoluteSourcePath.ToString(), targetTargetNodePath.ToString()))
      return None
    elif targetTargetNodeElement.GetType()==ResedaType.NODE:
      targetTargetNode = targetTargetNodeElement
    else:
      absoluteSourcePath = sourceEdge.GetAbsolutePath()
      self.RaiseError("In connection {0:s}, type mismatch for source node {1:s}\n".format(absoluteSourcePath.ToString(), targetTargetNodePath.ToString()))
      return None
    # 4. Creation or update of the connection
    if targetElement==None:
      if sourceEdge.GetType()==ResedaType.EDGE:
        targetEdge = self._targetModel.NewEdge(targetPath, targetSourceNodePath, targetTargetNodePath, targetContainer)
      else:
        targetEdge = self._targetModel.NewArc(targetPath, targetSourceNodePath, targetTargetNodePath, targetContainer)
    else:
      targetEdge = targetElement
      targetEdge.SetType(sourceEdge.GetType())
      targetEdge.SetSourceNodePath(targetSourceNodePath)
      targetEdge.SetTargetNodePath(targetTargetNodePath)
    targetEdge.SetSourceNode(targetSourceNode)
    targetEdge.SetTargetNode(targetTargetNode)
    targetSourceNode.AddOutEdge(targetEdge)
    targetTargetNode.AddInEdge(targetEdge)
    if targetEdge.GetType()==ResedaType.ARC:
      targetSourceNode.AddInEdge(targetEdge)
      targetTargetNode.AddOutEdge(targetEdge)
    # 5. Extension of the attribute list
    actualTargetContainer = targetEdge.GetContainer()
    for name in sourceEdge.GetAttributeNames():
      sourceExpression = sourceEdge.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, targetContainer, actualTargetContainer)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetEdge.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateVariable(self, action):
    sourceVariable = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceVariable.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        targetVariable = self._targetModel.NewVariable(targetPath, targetContainer)
      else:
        absoluteSourcePath = sourceVariable.GetAbsolutePath()
        self.RaiseError("Variable {0:s} declared outside of its parent block\n".format(absoluteSourcePath.ToString()))
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==ResedaType.VARIABLE:
      targetVariable = targetElement
    else:
      absoluteSourcePath = sourceVariable.GetAbsolutePath()
      self.RaiseError("Type mismatch for variable {0:s}\n".format(absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    actualTargetContainer = targetVariable.GetContainer()
    sourceDefinition = sourceVariable.GetDefinition()
    targetDefinition = self._ContextualizedCopyOfSetExpression(sourceDefinition, targetContainer, actualTargetContainer)
    if targetDefinition==None:
      return S2MLFlattenerAction.FAILED
    targetVariable.SetDefinition(targetDefinition)
    return S2MLFlattenerAction.COMPLETED

  def _ContextualizedCopyOfSetExpression(self, expression, targetContainer, actualTargetContainer):
    if expression.GetType()==ResedaSetExpression.ELEMENT_REFERENCE:
      sourceReferencePath = expression.GetReferencePath()
      if sourceReferencePath.IsIdentifier():
        identifier = sourceReferencePath.ToIdentifier()
        argument = self.GetValueOfIdentifierAsExpression(identifier)
        if argument!=None:
          environment = self._environment
          self._environment = None
          targetExpression = self._ContextualizedCopyOfSetExpression(argument, targetContainer, actualTargetContainer)
          self._environment = environment
          return targetExpression
      targetElement = targetContainer.AccessNamedElement(sourceReferencePath)
      if targetElement==None: # should not happen
        targetReferencePath = S2MLPath.Copy(sourceReferencePath)
      elif sourceReferencePath.IsAbsolute():
        targetReferencePath = targetElement.GetContextualizedAbsolutePath()
      else:
        targetReferencePath = actualTargetContainer.GetRelativePathToOtherElement(targetElement)
      return ResedaSetExpression.NewElementReference(targetReferencePath)
    if expression.GetType()==ResedaSetExpression.SELECTION:
      child = expression.GetChild(0);
      targetChild = self._ContextualizedCopyOfSetExpression(child, targetContainer, actualTargetContainer)
      if targetChild==None:
        return None
      testOnAttributes = expression.GetTestOnAttributes(0)
      targetTestOnAttributes = testOnAttributes.Copy()
      return ResedaSetExpression.NewSelection(targetChild, targetTestOnAttributes)
    if expression.GetType()==ResedaSetExpression.REACHABLE_NODES:
      child = expression.GetChild(0);
      targetChild = self._ContextualizedCopyOfSetExpression(child, targetContainer, actualTargetContainer)
      if targetChild==None:
        return None
      testOnAttributesOfNodes = expression.GetTestOnAttributes(0)
      targetTestOnAttributesOfNodes = testOnAttributesOfNodes.Copy()
      testOnAttributesOfEdges = expression.GetTestOnAttributes(1)
      targetTestOnAttributesOfEdges = testOnAttributesOfEdges.Copy()
      return ResedaSetExpression.NewReachableNodesOperation(ResedaSetExpression.REACHABLE_NODES, targetChild, testOnAttributesOfNodes, testOnAttributesOfEdges)
    if expression.GetType()==ResedaSetExpression.COREACHABLE_NODES:
      child = expression.GetChild(0);
      targetChild = self._ContextualizedCopyOfSetExpression(child, targetContainer, actualTargetContainer)
      if targetChild==None:
        return None
      testOnAttributesOfNodes = expression.GetTestOnAttributes(0)
      targetTestOnAttributesOfNodes = testOnAttributesOfNodes.Copy()
      testOnAttributesOfEdges = expression.GetTestOnAttributes(1)
      targetTestOnAttributesOfEdges = testOnAttributesOfEdges.Copy()
      return ResedaSetExpression.NewReachableNodesOperation(ResedaSetExpression.COREACHABLE_NODES, targetChild, testOnAttributesOfNodes, testOnAttributesOfEdges)
    childrenCopy = self._ContextualizedCopyOfChildrenOfSetExpression(expression, targetContainer, actualTargetContainer)
    if childrenCopy==None:
      return None
    expressionCopy = ResedaSetExpression.NewOperation(expression.GetType(), childrenCopy)
    return expressionCopy

  def _ContextualizedCopyOfChildrenOfSetExpression(self, expression, targetContainer, actualTargetContainer):
    childrenCopy = []
    for child in expression.GetChildren():
      childCopy = self._ContextualizedCopyOfSetExpression(child, targetContainer, actualTargetContainer)
      if childCopy==None:
        return None
      childrenCopy.append(childCopy)
    return childrenCopy

# 4) Flattening of models
# -----------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if sourceElement.GetType()==ResedaType.NODE:
      status = self._FlattenNode(action)
    elif sourceElement.GetType()==ResedaType.EDGE:
      status = self._FlattenEdge(action)
    elif sourceElement.GetType()==ResedaType.ARC:
      status = self._FlattenEdge(action)
    elif sourceElement.GetType()==ResedaType.VARIABLE:
      status = self._FlattenVariable(action)
    else:
      status = S2MLFlattener.FlattenElement(self, action)
    return status

  def _FlattenNode(self, action):
    sourceContainer = action.GetDatum(0)
    sourceNode = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceNode.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetNode = targetContainer.AccessNamedElement(targetPath)
    if targetNode==None:
      targetNode = self._targetModel.NewNode(targetPath, targetContainer)
    for name in sourceNode.GetAttributeNames():
      sourceExpression = sourceNode.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetNode.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenEdge(self, action):
    sourceContainer = action.GetDatum(0)
    sourceEdge = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceEdge.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetEdge = targetContainer.AccessNamedElement(targetPath)
    sourceSourceNode = sourceEdge.GetSourceNode()
    targetSourceNodePath = sourceSourceNode.GetAbsolutePath()
    targetSourceNodePath.RemoveFirstIdentifier()
    targetSourceNodePath.MakeIdentifier()
    targetSourceNode = targetContainer.AccessNamedElement(targetSourceNodePath)
    sourceTargetNode = sourceEdge.GetTargetNode()
    targetTargetNodePath = sourceTargetNode.GetAbsolutePath()
    targetTargetNodePath.RemoveFirstIdentifier()
    targetTargetNodePath.MakeIdentifier()
    targetTargetNode = targetContainer.AccessNamedElement(targetTargetNodePath)
    if targetEdge==None:
      if sourceEdge.GetType()==ResedaType.EDGE:
        targetEdge = self._targetModel.NewEdge(targetPath, targetSourceNodePath, targetTargetNodePath, targetContainer)
      else:
        targetEdge = self._targetModel.NewArc(targetPath, targetSourceNodePath, targetTargetNodePath, targetContainer)
    else:
      targetEdge.SetSourceNodePath(targetSourceNodePath)
      targetEdge.SetTargetNodePath(targetTargetNodePath)
    targetEdge.SetSourceNode(targetSourceNode)
    targetEdge.SetTargetNode(targetTargetNode)
    targetSourceNode.AddOutEdge(targetEdge)
    targetTargetNode.AddInEdge(targetEdge)
    if targetEdge.GetType()==ResedaType.ARC:
      targetSourceNode.AddInEdge(targetEdge)
      targetTargetNode.AddOutEdge(targetEdge)
    for name in sourceEdge.GetAttributeNames():
      sourceExpression = sourceEdge.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetEdge.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenVariable(self, action):
    sourceContainer = action.GetDatum(0)
    sourceSet = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceSet.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetVariable = targetContainer.AccessNamedElement(targetPath)
    if targetVariable==None:
      targetVariable = self._targetModel.NewVariable(targetPath, targetContainer)
    sourceDefinition = sourceSet.GetDefinition()
    targetDefinition= self._FlattenSetExpression(sourceContainer, sourceDefinition)
    if targetDefinition==None:
      return S2MLFlattenerAction.FAILED
    targetVariable.SetDefinition(targetDefinition)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenSetExpression(self, sourceContainer, sourceExpression):
    if sourceExpression.GetType()==ResedaSetExpression.ELEMENT_REFERENCE:
      sourceReferencePath = sourceExpression.GetReferencePath()
      sourceReferredElement = sourceContainer.AccessNamedElement(sourceReferencePath)
      #if sourceReferredElement==None or not S2MLModel.IsParameter(sourceReferredElement):
      if sourceReferredElement==None:
        absoluteSourcePath = sourceContainer.GetAbsolutePath()
        message = "In %s, undefined reference %s\n" % (absoluteSourcePath.ToString(), sourceReferencePath.ToString())
        self.RaiseError(message)
        return None
      targetReferencePath = sourceReferredElement.GetAbsolutePath()
      targetReferencePath.RemoveFirstIdentifier()
      targetReferencePath.MakeIdentifier()
      return ResedaSetExpression.NewElementReference(targetReferencePath)
    if expression.GetType()==ResedaSetExpression.SELECTION:
      child = expression.GetChild(0);
      targetChild = self._FlattenSetExpression(sourceContainer, child)
      if targetChild==None:
        return None
      testOnAttributes = expression.GetTestOnAttributes(0)
      targetTestOnAttributes = testOnAttributes.Copy()
      return ResedaSetExpression.NewSelection(targetChild, targetTestOnAttributes)
    if expression.GetType()==ResedaSetExpression.REACHABLE_NODES:
      child = expression.GetChild(0);
      targetChild = self._FlattenSetExpression(sourceContainer, child)
      if targetChild==None:
        return None
      testOnAttributesOfNodes = expression.GetTestOnAttributes(0)
      targetTestOnAttributesOfNodes = testOnAttributesOfNodes.Copy()
      testOnAttributesOfEdges = expression.GetTestOnAttributes(1)
      targetTestOnAttributesOfEdges = testOnAttributesOfEdges.Copy()
      return ResedaSetExpression.NewReachableNodesOperation(ResedaSetExpression.REACHABLE_NODES, targetChild, targetTestOnAttributesOfNodes, testOnAttributesOfEdges)
    if expression.GetType()==ResedaSetExpression.COREACHABLE_NODES:
      child = expression.GetChild(0);
      targetChild = self._FlattenSetExpression(sourceContainer, child)
      if targetChild==None:
        return None
      testOnAttributesOfNodes = expression.GetTestOnAttributes(0)
      targetTestOnAttributesOfNodes = testOnAttributesOfNodes.Copy()
      testOnAttributesOfEdges = expression.GetTestOnAttributes(1)
      targetTestOnAttributesOfEdges = testOnAttributesOfEdges.Copy()
      return ResedaSetExpression.NewReachableNodesOperation(ResedaSetExpression.COREACHABLE_NODES, targetChild, targetTestOnAttributesOfNodes, testOnAttributesOfEdges)
    targetChildren = self._FlattenChildrenOfSetExpression(sourceContainer, sourceExpression)
    if targetChildren==None:
      return None
    targetExpression = ResedaSetExpression.NewOperation(sourceExpression.GetType(), targetChildren)
    return targetExpression
  
  def _FlattenChildrenOfSetExpression(self, sourceContainer, sourceExpression):
    targetChildren = []
    for sourceChild in sourceExpression.GetChildren():
      targetChild = self._FlattenSetExpression(sourceContainer, sourceChild)
      if targetChild==None:
        return None
      targetChildren.append(targetChild)
    return targetChildren

