# Reseda version 1.0.0
# ResedaSimulator.py
# by Antoine Rauzy
# Copyrights (c) 2018-2021 AltaRica Association
#
# This file is part of Reseda.
# Reseda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Reseda is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Reseda.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Parameters
# 4) Simulation
# 5) Result matrix

# 1) Imported modules
# -------------------

import time

from S2MLCore import *

from ResedaModels import *

# 2) Creator
# ----------

class ResedaSimulator:

  def __init__(self, engine):
    from ResedaEngine import ResedaEngine
    self._engine = engine
    self._expressionManager = self._engine.GetExpressionManager()
    self._expressionManager.SetCalculateValueOfElement(ResedaSimulator.CalculateValueOfElement)
    self._targetBlock = None
    self._parameters = []
    self._variables = []
    self._equations = []
    self._observers = []
    self._matrix = None
    self._seed = 12345
    self._numberOfTries = 0
    self._tryIndex = 0
    self._outputFileName = None
    self._outputFileOpeningMode = "write"
    self._missionTimes = []
    self._missionTimeIndex = 0
    self._executionTime = 0

  def RaiseError(self, message):
    self._engine.RaiseError(message)

# 3) Parameters
# -------------

  def GetSeed(self):
    return self._seed
  def SetSeed(self, value):
    self._seed = value

  def GetNumberOfTries(self):
    return self._numberOfTries
  def SetNumberOfTries(self, numberOfTries):
    self._numberOfTries = numberOfTries

  def GetOutputFileName(self):
    return self._outputFileName
  def SetOutputFileName(self, outputFileName):
    self._outputFileName = outputFileName

  def GetOutputFileOpeningMode(self):
    return self._outputFileOpeningMode
  def SetOutputFileOpeningMode(self, openingMode):
    self._outputFileOpeningMode = openingMode

# 4) Simulation
# -------------

  def CalculateValuesOfElements(self, targetBlock, missionTimes):
    self._targetBlock = targetBlock
    self._executionTime = time.time()
    self._missionTimes.clear()
    self._missionTimes.extend(missionTimes)
    self._missionTimes.sort()
    self._CollectModelElements()
    self._missionTimeIndex = -1
    self._ResetValuesOfElements()
    self._ResetAbscissasOfRandomDeviatesOfElements()
    for missionTime in self._missionTimes:
      self._missionTimeIndex += 1
      self._expressionManager.SetMissionTime(missionTime)
      self._ResetValuesOfElements()
      self._CalculateValuesOfElements()
    self._executionTime = time.time() - self._executionTime
    self._CreateResultMatrix()
    self._PrintResultMatrix()
    return 0

  def PerformStochasticSimulation(self, targetBlock, missionTimes):
    from ResedaEngine import ResedaEngine
    self._targetBlock = targetBlock
    self._executionTime = time.time()
    self._missionTimes.clear()
    self._missionTimes.extend(missionTimes)
    self._missionTimes.sort()
    self._CollectModelElements()
    random.seed(self._seed)
    for self._tryIndex in range(0, self._numberOfTries):
      if self._engine.GetExecutionStatus()!=S2MLEngine.ACTIVE:
        return 1
      self._missionTimeIndex = -1
      self._ResetValuesOfElements()
      self._DrawAbscissasOfRandomDeviatesOfElements()
      for missionTime in self._missionTimes:
        if self._engine.GetExecutionStatus()!=S2MLEngine.ACTIVE:
          return 1
        self._missionTimeIndex += 1
        self._expressionManager.SetMissionTime(missionTime)
        self._ResetValuesOfElements()
        self._CalculateValuesOfElements()
    self._executionTime = time.time() - self._executionTime
    self._CreateResultMatrix()
    self._PrintResultMatrix()
    return 0

  def _CollectModelElements(self):
    self._parameters.clear()
    self._variables.clear()
    self._equations.clear()
    self._observers.clear()
    for element in self._targetBlock.GetElements():
      if element.GetType()==ResedaType.PARAMETER:
        self._parameters.append(element)
      elif element.GetType()==ResedaType.VARIABLE:
        self._variables.append(element)
      elif element.GetType()==ResedaType.ASSIGNMENT:
        self._equations.append(element)
      elif element.GetType()==ResedaType.OBSERVER:
        self._observers.append(element)
        element.CreateIndicators(len(self._missionTimes))

  @staticmethod
  def CalculateValueOfElement(expressionManager, element):
    if element.GetType()==ResedaType.PARAMETER:
     return expressionManager.CalculateValueOfParameter(element)
    if element.GetType()==ResedaType.VARIABLE:
      return ResedaSimulator.CalculateValueOfVariable(expressionManager, element)
    if element.GetType()==ResedaType.OBSERVER:
      return ResedaSimulator.CalculateValueOfObserver(expressionManager, element)
    return S2MLConstant.NewNone()

  @staticmethod
  def CalculateValueOfVariable(expressionManager, variable):
    return variable.GetValue()

  @staticmethod
  def CalculateValueOfObserver(expressionManager, observer):
    if observer.IsValueCalculated():
      value = observer.GetValue()
    else:
      definition = observer.GetDefinition()
      value = expressionManager.CalculateValueOfExpression(definition)
      observer.SetValue(value)
    return observer.GetValue()

  def _ResetValuesOfElements(self):
    for parameter in self._parameters:
      parameter.ResetValue()
    for variable in self._variables:
      variable.ResetValue()
    for observer in self._observers:
      observer.ResetValue()

  def _CalculateValuesOfElements(self):
    for parameter in self._parameters:
      self._expressionManager.CalculateValueOfParameter(parameter)
    for equation in self._equations:
      self._CalculateValueOfVariablesFromEquation(equation)
    for observer in self._observers:
      self._CalculateValueOfObserver(observer)

  def _CalculateValueOfVariablesFromEquation(self, equation):
    leftMember = equation.GetLeftMember()
    variable = leftMember.GetReferredElement()
    rightMember = equation.GetRightMember()
    value = self._expressionManager.CalculateValueOfExpression(rightMember)
    variable.SetValue(value)
    # print(variable.GetIdentifier() + " " + str(value.GetValue()))

  def _CalculateValueOfObserver(self, observer):
    value = ResedaSimulator.CalculateValueOfObserver(self._expressionManager, observer)
    if self._missionTimeIndex<0 or self._missionTimeIndex>=len(self._missionTimes):
      return
    observer.RecordValueInIndicator(self._missionTimeIndex)

  def _ResetAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(definition)
    for equation in self._equations:
      rightMember = equation.GetRightMember()
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(rightMember)
    for observer in self._observers:
      definition = observer.GetDefinition()
      self._expressionManager.ResetAbscissasOfRandomDeviatesInExpression(definition)

  def _DrawAbscissasOfRandomDeviatesOfElements(self):
    for parameter in self._parameters:
      definition = parameter.GetDefinition()
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(definition)
    for equation in self._equations:
      rightMember = equation.GetRightMember()
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(rightMember)
    for observer in self._observers:
      definition = observer.GetDefinition()
      self._expressionManager.DrawAbscissasOfRandomDeviatesInExpression(definition)


# 4) Management of result matrix
# ------------------------------

  def _CreateResultMatrix(self):
    self._matrix = S2MLMatrix()
    row = self._matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Model"))
    row.Put(1, S2MLConstant.NewString(self._targetBlock.GetIdentifier()))
    if self._numberOfTries>0:
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Number-of-tries"))
      row.Put(1, S2MLConstant.NewInteger(self._numberOfTries))
      row = self._matrix.AppendRow()
      row.Put(0, S2MLConstant.NewString("Execution-time"))
      hours, remainder = divmod(self._executionTime, 3600)
      minutes, seconds = divmod(remainder, 60)
      timeString = "{:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds)
      row.Put(1, S2MLConstant.NewString(timeString))
    for observer in self._observers:
      self._CreateResultRowsForObserver(observer)

  def _CreateResultRowsForObserver(self, observer):
    row = self._matrix.AppendRow()
    row.Put(0, S2MLConstant.NewString("Observer"))
    row.Put(1, S2MLConstant.NewString(observer.GetIdentifier()))
    missionTimeIndex = -1
    for missionTime in self._missionTimes:
      missionTimeIndex += 1
      row = self._matrix.AppendRow()
      row.Put(1, S2MLConstant.NewString("Mission-time"))
      row.Put(2, S2MLConstant.NewReal(missionTime))
      indicator = observer.GetIndicator(missionTimeIndex)
      self._CreateResultRowsForIndicator(indicator)

  def _CreateResultRowsForIndicator(self, indicator):
      self._CreateHeaderRowForIndicator(indicator)
      domain = indicator.GetDomain()
      if domain.IsNumerical():
        self._CreateResultRowsForRealIndicator(indicator)
      elif domain.IsBoolean():
        self._CreateResultRowsForBooleanIndicator(indicator)
      elif domain.IsEnumeration():
        self._CreateResultRowsForEnumerationIndicator(indicator)

  def _CreateResultRowsForRealIndicator(self, indicator):
    profile = indicator.GetProfile()
    bin = indicator.GetMeasure()
    row = self._CreateRowForMeasure(indicator, bin)
    if profile.GetQuantiles() and self._numberOfTries>0:
      row1 = self._matrix.AppendRow()
      row2 = self._matrix.AppendRow()
      column = 2
      row1.Put(column, S2MLConstant.NewString("Quantiles"))
      quantiles = indicator.ExtractQuantiles()
      for (quantile, value) in quantiles:
        column += 1
        row1.Put(column, S2MLConstant.NewReal(quantile))
        row2.Put(column, S2MLConstant.NewReal(value))
    if profile.GetDistribution() and self._numberOfTries>0:
      row1 = self._matrix.AppendRow()
      row2 = self._matrix.AppendRow()
      column = 2
      row1.Put(column, S2MLConstant.NewString("Distribution"))
      points = indicator.ExtractDistribution()
      for (x, y) in points:
        column += 1
        row1.Put(column, S2MLConstant.NewReal(x))
        row2.Put(column, S2MLConstant.NewReal(y))
    if profile.GetCumulativeDistribution() and self._numberOfTries>0:
      row1 = self._matrix.AppendRow()
      row2 = self._matrix.AppendRow()
      column = 2
      row1.Put(column, S2MLConstant.NewString("Cumulative-distribution"))
      points = indicator.ExtractCumulativeDistribution()
      for (x, y) in points:
        column += 1
        row1.Put(column, S2MLConstant.NewReal(x))
        row2.Put(column, S2MLConstant.NewReal(y))
    if profile.GetBootstrapping() and self._numberOfTries>0:
      row = self._matrix.AppendRow()
      row.Put(2, S2MLConstant.NewString("Bootstrapping"))
      row.Put(3, S2MLConstant.NewInteger(indicator.GetNumberOfSamples()))
      row.Put(4, S2MLConstant.NewReal(indicator.GetSamplingRatio()))
      indicator.CalculateBootstrappedMeasures()
      bin = indicator.GetBootstrappedMeasure()
      row = self._CreateRowForMeasure(indicator, bin)

  def _CreateResultRowsForBooleanIndicator(self, indicator):
    profile = indicator.GetProfile()
    bin = indicator.GetMeasure()
    row = self._CreateRowForMeasure(indicator, bin)
    if profile.GetBootstrapping():
      row = self._matrix.AppendRow()
      row.Put(2, S2MLConstant.NewString("Bootstrapping"))
      row.Put(3, S2MLConstant.NewInteger(indicator.GetNumberOfSamples()))
      row.Put(4, S2MLConstant.NewReal(indicator.GetSamplingRatio()))
      indicator.CalculateBootstrappedMeasures()
      bin = indicator.GetBootstrappedMeasure()
      row = self._CreateRowForMeasure(indicator, bin)

  def _CreateResultRowsForEnumerationIndicator(self, indicator):
    profile = indicator.GetProfile()
    domain = indicator.GetDomain()
    for index in range(0, domain.GetSize()):
      bin = indicator.GetMeasure(index)
      row = self._CreateRowForMeasure(indicator, bin)
      row.Put(2, domain.GetConstant(index))
    if profile.GetBootstrapping():
      row = self._matrix.AppendRow()
      row.Put(2, S2MLConstant.NewString("Bootstrapping"))
      row.Put(3, S2MLConstant.NewInteger(indicator.GetNumberOfSamples()))
      row.Put(4, S2MLConstant.NewReal(indicator.GetSamplingRatio()))
      for index in range(0, domain.GetSize()):
        indicator.CalculateBootstrappedMeasures()
        bin = indicator.GetBootstrappedMeasure(index)
        row = self._CreateRowForMeasure(indicator, bin)
        row.Put(2, domain.GetConstant(index))

  def _CreateHeaderRowForIndicator(self, indicator):
    profile = indicator.GetProfile()
    domain = indicator.GetDomain()
    row = self._matrix.AppendRow()
    column = 2
    if self._numberOfTries==0:
      column += 1
      row.Put(column, S2MLConstant.NewString("Value"))
      return row
    if profile.GetMean():
      column += 1
      row.Put(column, S2MLConstant.NewString("Mean"))
    if profile.GetStandardDeviation():
      column += 1
      row.Put(column, S2MLConstant.NewString("Standard-deviation"))
    if profile.GetConfidenceInterval90():
      column += 1
      row.Put(column, S2MLConstant.NewString("90% confidence interval"))
      column += 1
    if profile.GetConfidenceInterval95():
      column += 1
      row.Put(column, S2MLConstant.NewString("95% confidence interval"))
      column += 1
    if profile.GetConfidenceInterval99():
      column += 1
      row.Put(column, S2MLConstant.NewString("99% confidence interval"))
      column += 1
    if profile.GetExtrema():
      column += 1
      row.Put(column, S2MLConstant.NewString("Minimum"))
      column += 1
      row.Put(column, S2MLConstant.NewString("Maximum"))
    return row

  def _CreateRowForMeasure(self, indicator, bin):
    profile = indicator.GetProfile()
    domain = indicator.GetDomain()
    mean = bin.GetMean()
    row = self._matrix.AppendRow()
    column = 2
    if self._numberOfTries==0:
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean))
      return row
    if profile.GetMean():
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean))
    if profile.GetStandardDeviation():
      standardDeviation = bin.GetStandardDeviation()
      column += 1
      row.Put(column, S2MLConstant.NewReal(standardDeviation))
    if profile.GetConfidenceInterval90():
      slack = bin.GetErrorFactor90()
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean - slack))
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean + slack))
    if profile.GetConfidenceInterval95():
      slack = bin.GetErrorFactor95()
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean - slack))
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean + slack))
    if profile.GetConfidenceInterval99():
      slack = bin.GetErrorFactor99()
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean - slack))
      column += 1
      row.Put(column, S2MLConstant.NewReal(mean + slack))
    if profile.GetExtrema():
      minimum = bin.GetMinimumValue()
      maximum = bin.GetMaximumValue()
      column += 1
      row.Put(column, S2MLConstant.NewReal(minimum))
      column += 1
      row.Put(column, S2MLConstant.NewReal(maximum))
    return row

  def _PrintResultMatrix(self):
    printer = self._engine.GetPrinter()
    if self._outputFileName==None:
      printer.PrintMatrix(self._matrix, self._engine.GetOutputFile())
    else:
      filePath = self._engine.GetFilePath(self._outputFileName)
      printer.ExportMatrix(self._matrix, filePath, self._outputFileOpeningMode[0])